# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'DAQ_Measurement_GUI.ui'
#
# Created by: qtpy UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from qtpy import QtCore, QtGui, QtWidgets
from pymodaq.resources.QtDesigner_Ressources import QtDesigner_ressources_rc


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(754, 326)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.splitter_2 = QtWidgets.QSplitter(Form)
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.setObjectName("splitter_2")
        self.splitter = QtWidgets.QSplitter(self.splitter_2)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.gridLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.gridLayoutWidget.setObjectName("gridLayoutWidget")
        self.gridLayout = QtWidgets.QGridLayout(self.gridLayoutWidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.formula_edit = QtWidgets.QPlainTextEdit(self.gridLayoutWidget)
        self.formula_edit.setMaximumSize(QtCore.QSize(16777215, 50))
        self.formula_edit.setPlainText("")
        self.formula_edit.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard | QtCore.Qt.TextSelectableByMouse)
        self.formula_edit.setObjectName("formula_edit")
        self.gridLayout.addWidget(self.formula_edit, 4, 1, 1, 1)
        self.Quit_pb = QtWidgets.QPushButton(self.gridLayoutWidget)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/close2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.Quit_pb.setIcon(icon)
        self.Quit_pb.setObjectName("Quit_pb")
        self.gridLayout.addWidget(self.Quit_pb, 1, 0, 1, 2)
        self.measurement_type_combo = QtWidgets.QComboBox(self.gridLayoutWidget)
        self.measurement_type_combo.setObjectName("measurement_type_combo")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.gridLayout.addWidget(self.measurement_type_combo, 2, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_11.setAlignment(QtCore.Qt.AlignCenter)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 3, 0, 1, 1)
        self.label = QtWidgets.QLabel(self.gridLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Tahoma")
        font.setPointSize(14)
        font.setBold(True)
        font.setItalic(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 2)
        self.measure_subtype_combo = QtWidgets.QComboBox(self.gridLayoutWidget)
        self.measure_subtype_combo.setObjectName("measure_subtype_combo")
        self.measure_subtype_combo.addItem("")
        self.measure_subtype_combo.addItem("")
        self.measure_subtype_combo.addItem("")
        self.gridLayout.addWidget(self.measure_subtype_combo, 3, 1, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_9.setAlignment(QtCore.Qt.AlignCenter)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 2, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_7.setAlignment(QtCore.Qt.AlignCenter)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 4, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 2, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 5, 0, 1, 1)
        self.result_sb = QtWidgets.QDoubleSpinBox(self.gridLayoutWidget)
        font = QtGui.QFont()
        font.setPointSize(8)
        self.result_sb.setFont(font)
        self.result_sb.setReadOnly(True)
        self.result_sb.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        self.result_sb.setMinimum(-100000000.0)
        self.result_sb.setMaximum(100000000.0)
        self.result_sb.setObjectName("result_sb")
        self.gridLayout.addWidget(self.result_sb, 5, 1, 1, 1)
        self.horizontalLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.horizontalLayoutWidget.setObjectName("horizontalLayoutWidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.horizontalLayoutWidget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.result_lcd = QtWidgets.QLCDNumber(self.horizontalLayoutWidget)
        self.result_lcd.setSmallDecimalPoint(True)
        self.result_lcd.setDigitCount(10)
        self.result_lcd.setObjectName("result_lcd")
        self.horizontalLayout.addWidget(self.result_lcd)
        self.gridLayout_2.addWidget(self.splitter_2, 0, 0, 1, 1)
        self.StatusBarLayout = QtWidgets.QHBoxLayout()
        self.StatusBarLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.StatusBarLayout.setObjectName("StatusBarLayout")
        self.gridLayout_2.addLayout(self.StatusBarLayout, 1, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.Quit_pb.setText(_translate("Form", "Quit"))
        self.measurement_type_combo.setItemText(0, _translate("Form", "Cursor Intensity Integration"))
        self.measurement_type_combo.setItemText(1, _translate("Form", "Max"))
        self.measurement_type_combo.setItemText(2, _translate("Form", "Min"))
        self.measurement_type_combo.setItemText(3, _translate("Form", "Gaussian Fit"))
        self.measurement_type_combo.setItemText(4, _translate("Form", "Lorentzian Fit"))
        self.measurement_type_combo.setItemText(5, _translate("Form", "Exponential Decay Fit"))
        self.label_11.setText(_translate("Form", "Measurement SubType: "))
        self.label.setText(_translate("Form", "DAQ Measurement"))
        self.measure_subtype_combo.setItemText(0, _translate("Form", "dx"))
        self.measure_subtype_combo.setItemText(1, _translate("Form", "amplitude"))
        self.measure_subtype_combo.setItemText(2, _translate("Form", "x0"))
        self.label_9.setText(_translate("Form", "Measurement Type: "))
        self.label_7.setText(_translate("Form", "Formula\n"
                                                "(the variable is \"x\"):"))
        self.label_4.setText(_translate("Form", "Result:"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())
